@@ -2,6 +2,23 @@
 Parse ASCII PSF Files
 """
 
+# License {{{1
+# Copyright (C) 2016-2023 Kenneth S. Kundert
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program.  If not, see http://www.gnu.org/licenses/.
+
+
 # Imports {{{1
 import ply.lex
 import ply.yacc
@@ -37,6 +54,10 @@ class Value(Info):
     pass
 
 
+class Array(Info):
+    pass
+
+
 # Exceptions {{{1
 # ParseError {{{2
 class ParseError(Exception):
@@ -110,6 +131,7 @@ class TokenLocation(object):
 # Lexer {{{1
 # List of the token names
 reserved = {rw: rw for rw in [
+    'ARRAY',
     'BYTE',
     'COMPLEX',
     'DOUBLE',
@@ -121,6 +143,7 @@ reserved = {rw: rw for rw in [
     'LONG',
     'NAN',
     'PROP',
+    'STRING',
     'STRUCT',
     'SWEEP',
     'TRACE',
@@ -130,20 +153,24 @@ reserved = {rw: rw for rw in [
 tokens = [
     'INTEGER',
     'REAL',
-    'STRING',
+    'QUOTED_STRING',
 ] + list(reserved.values())
 
 # Literal tokens
-literals = r'()'
+literals = r'()*'
 
 # Regular expression tokens
 # Regular expressions that define numbers
 t_INTEGER = r"-?[0-9]+"
 t_REAL = r"[+-]?[0-9]+\.[0-9]*([eE][+-][0-9]+)?"
-t_NAN = r"nan|NaN"
+t_NAN = r"nan|NaN|inf"
 
 # Regular expression for a string
-t_STRING = r'"[^\n"]*"'
+t_QUOTED_STRING = r'"([^\\\n"]|(\\.))*"'
+    # The complexity is because of the case "He yelled \"You are a fool!\".".
+    # The first part says string cannot contain a backslash, newline, or a
+    # quote. The second case allows backslashes when combined with any other
+    # character, which allows \" and \\.
 
 
 # Identifiers
@@ -179,6 +206,11 @@ def p_contents_without_sweep(p):
     p[0] = (p[1], p[2], None, None, p[3])
 
 
+def p_contents_only_header(p):
+    "contents : header_section end"
+    p[0] = (p[1], {}, None, None, {})
+
+
 def p_header_section(p):
     "header_section : HEADER named_values"
     p[0] = dict(p[2])
@@ -219,10 +251,15 @@ def p_real_value(p):
 
 
 def p_string(p):
-    "string : STRING"
+    "string : QUOTED_STRING"
     p[0] = (p[1][1:-1]).replace('\\', '')
 
 
+def p_star(p):
+    "star : '*'"
+    p[0] = p[1]
+
+
 def p_type_section(p):
     "type_section : TYPE types"
     p[0] = dict(p[2])
@@ -273,8 +310,11 @@ def p_kind(p):
              | INT
              | BYTE
              | LONG
+             | STRING
+             | array
              | struct
              | prop
+             | star
     """
     p[0] = p[1]
 
@@ -284,6 +324,11 @@ def p_struct(p):
     p[0] = Struct(types=dict(p[3]))
 
 
+def p_array(p):
+    "array : ARRAY '(' star ')'"
+    p[0] = Array(members=p[3])
+
+
 def p_prop(p):
     "prop : PROP '(' named_values ')'"
     p[0] = dict(p[3])
@@ -408,6 +453,32 @@ def p_named_signal_with_type(p):
     p[0] = (p[1], p[2], p[3])
 
 
+def p_named_string_signal_with_type(p):
+    """
+    signal_value : string string string
+    """
+    p[0] = (p[1], p[2], p[3])
+
+
+def p_numbers(p):
+    "numbers : numbers number"
+    p[1].append(p[2])
+    p[0] = p[1]
+
+
+def p_last_number(p):
+    "numbers : number"
+    p[0] = [p[1]]
+
+
+def p_number(p):
+    """
+    number : simple_number
+           | composite_number
+    """
+    p[0] = p[1]
+
+
 def p_integer_number(p):
     "simple_number : INTEGER"
     p[0] = int(p[1])
@@ -424,44 +495,14 @@ def p_real_number(p):
     p[0] = float(p[1])
 
 
-def p_simple_numbers(p):
-    "simple_numbers : simple_numbers simple_number"
-    p[1].append(p[2])
-    p[0] = p[1]
-
-
-def p_simple_numbers_last(p):
-    "simple_numbers : simple_number"
-    p[0] = [p[1]]
-
-
 def p_composite_number(p):
     """
-    composite_number : '(' simple_numbers ')'
-                     | '(' simple_numbers ')' prop
+    composite_number : '(' numbers ')'
+                     | '(' numbers ')' prop
     """
     p[0] = tuple(p[2])
 
 
-def p_composite_numbers(p):
-    "composite_numbers : composite_numbers composite_number"
-    p[1].append(p[2])
-    p[0] = p[1]
-
-
-def p_composite_numbers_last(p):
-    "composite_numbers : composite_number"
-    p[0] = [p[1]]
-
-
-def p_numbers(p):
-    """
-    numbers : simple_numbers
-            | composite_numbers
-    """
-    p[0] = p[1]
-
-
 def p_end(p):
     "end : END"
 
