@@ -2,6 +2,23 @@
 Read PSF File
 """
 
+# License {{{1
+# Copyright (C) 2018-2023 Kenneth S. Kundert
+#
+# This program is free software: you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation, either version 3 of the License, or
+# (at your option) any later version.
+#
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+#
+# You should have received a copy of the GNU General Public License
+# along with this program.  If not, see [http://www.gnu.org/licenses/].
+
+
 # Imports {{{1
 from .parse import ParsePSF, ParseError
 from inform import Error, Info, join, log, os_error
@@ -34,6 +51,7 @@ unicode_unit_maps = {
     r'\bR\b': 'Ω',
     r'\bI\b': 'A',
     r'\bC\b': 'F',
+    r'\bDeg\b': '°',
 }
 
 
@@ -145,7 +163,12 @@ class PSF:
                         ordinate = ordinate,
                         type = t,
                         access = t.name,
-                        units = t.units,
+                        units = t.units if t.units else t.name,
+                            # I think that terminal currents are misnamed
+                            # they should be named "I" but they are named "A"
+                            # and "A" is a type where no units are specified,
+                            # so as a hack, use the name rather than
+                            # go without units
                         meta = meta,
                     )
                     signals[joined_name] = signal
@@ -158,7 +181,10 @@ class PSF:
                 if type.struct:
                     for t, v in zip(type.struct.types.values(), value.values[0][0]):
                         n = f'{name}.{t.name}'
-                        v = Quantity(v, unicode_units(t.units))
+                        if 'float' in t.kind:
+                            v = Quantity(v, unicode_units(t.units))
+                        elif 'complex' in t.kind:
+                            v = complex(v[0], v[1])
                         signal = Signal(
                             name = n,
                             ordinate = v,
@@ -168,7 +194,13 @@ class PSF:
                         )
                         signals[n] = signal
                 else:
-                    v = Quantity(value.values[0][0], unicode_units(type.units))
+                    if 'float' in type.kind:
+                        v = Quantity(value.values[0][0], unicode_units(type.units))
+                    elif 'complex' in type.kind:
+                        v = complex(value.values[0][0], value.values[0][1])
+                    else:
+                        v = value.values[0]
+
                     signal = Signal(
                         name = name,
                         ordinate = v,
